import json
import os
from datetime import datetime
from logging import getLogger
from pathlib import Path
from time import sleep
from typing import Optional

import requests
from flask import Flask
from flask_sqlalchemy import SQLAlchemy

from .models import CrawlerTask

logger = getLogger()


class BaseCrawlerAdapter():

    @property
    def default_tasks_path(self) -> Path:
        from app.config import app_path
        return app_path.joinpath('crawling', 'tasks')

    @property
    def crawler_create_task_url(self):
        return f'{self.crawler_url}/task/create'

    @property
    def crawler_run_task_url(self):
        return f'{self.crawler_url}/task/run'

    @property
    def crawler_task_status_url(self):
        return f'{self.crawler_url}/task/status'

    @property
    def crawler_data_url(self):
        return f'{self.crawler_url}/data/get'

    def __init__(self, app: Flask, db: SQLAlchemy, tasks_path: Path = None):
        self.app = app
        self.db = db
        self.tasks_path = tasks_path if tasks_path else self.default_tasks_path

        self.crawler_url = app.config.get('CRAWLER_URL')

        logger.debug(f'CrawlerAdapter created')

        if app.config.get('CRAWLER_ADD_TASKS'):
            self.__add_tasks()
            logger.debug(f'Tasks from {self.tasks_path} added')

    def __add_tasks(self) -> None:
        logger.debug(f'Scan {self.tasks_path} directory')

        with os.scandir(self.tasks_path) as entries:
            entry: os.DirEntry
            for entry in entries:
                with open(entry, mode='r', encoding='utf8') as file:
                    task_body = json.load(file)
                    task_name = task_body.get('name')
                    if not task_name:
                        logger.error(f'Cannot import task from file {file.name}. Specify task name in task.name')
                        continue

                    self.__create_crawler_task(task_name=task_name, task_body=task_body)

    def __create_crawler_task(self, task_name: str, task_body: str) -> None:
        response = requests.post(self.crawler_create_task_url, json=task_body)

        if response.status_code != 200:
            logger.warning(f'Response status is {response.status_code}: Payload: {response.text}')

        data = response.json()
        if data:
            task_id = data['result']['id']

            crawler_task: CrawlerTask = CrawlerTask.query.filter_by(name=task_name).first()
            if not crawler_task:
                crawler_task = CrawlerTask()
                crawler_task.name = task_name
                self.db.session.add(crawler_task)

            crawler_task.task_id = task_id
            crawler_task.task_body = json.dumps(task_body)
            self.db.session.commit()

            logger.debug(f'Task {task_name} added')
        else:
            logger.warning('Response data is empty')

    def __run_task(self, crawler_task: CrawlerTask) -> None:
        task = json.loads(crawler_task.task_body)
        task['task_id'] = crawler_task.task_id

        response = requests.post(self.crawler_run_task_url, json=task)

        if response.status_code != 200:
            logger.error(f'Run task is not succeed. Status {response.status_code}. Payload: {response.text}')

        logger.debug(response.text)

    def __wait_until_task_finish(self, crawler_task: CrawlerTask, timeout: int) -> None:
        data = {
            'task_id': crawler_task.task_id
        }

        while True:
            response = requests.post(self.crawler_task_status_url, json=data)

            if response.status_code != 200:
                logger.warning(
                    f'Check task status is not succeed. Status {response.status_code}. Payload: {response.text}')
            logger.debug(response.text)

            status = response.json()['result']['status']
            logger.debug(f'{crawler_task.name} task status: {status}')
            if status == 'in_db':
                break

            sleep(timeout)

    def __fetch(self, crawler_task: CrawlerTask, limit: int, last_crawled: Optional[int] = None) -> dict:
        task = {
            'task_id': crawler_task.task_id,
            'start_ts': last_crawled if last_crawled is not None else crawler_task.last_crawled
        }
        if limit:
            task['limit'] = limit

        response = requests.post(self.crawler_data_url, json=task)

        data = response.json()['result']['data']

        logger.info(f'Crawled {len(data)} pieces of data')
        ctimes = [item.get('ctime') for item in data if item.get('ctime')]
        if ctimes:
            last_crawled = max(ctimes)
            logger.debug(f'last_crawled is {str(datetime.fromtimestamp(last_crawled))}')
            if last_crawled:
                crawler_task.last_crawled = last_crawled

        self.db.session.add(crawler_task)
        self.db.session.commit()

        return data

    def data(self, task_name: str, limit=None, timeout: int = 15, last_crawled: Optional[int] = None,
             run_task: Optional[bool] = None, wait_until_task_finish: Optional[bool] = None) -> dict:
        crawler_task: CrawlerTask = CrawlerTask.query.filter_by(name=task_name).first()
        if not crawler_task:
            raise KeyError(f'task {task_name} does not exist')

        logger.debug(f"run_task: parameter is {run_task}, config is {self.app.config.get('CRAWLER_RUN_TASK')}")

        if run_task is True or \
                (run_task is None and self.app.config.get('CRAWLER_RUN_TASK') is True):
            logger.debug('running __run_task')
            self.__run_task(crawler_task=crawler_task)

        logger.debug(f"wait_until_task_finish: parameter is {wait_until_task_finish}, "
                     f"config is {self.app.config.get('CRAWLER_WAIT_UNTIL_TASK_FINISH')}")

        if wait_until_task_finish is True or \
                (wait_until_task_finish is None and self.app.config.get('CRAWLER_WAIT_UNTIL_TASK_FINISH') is True):
            logger.debug('running __wait_until_task_finish')
            self.__wait_until_task_finish(crawler_task=crawler_task, timeout=timeout)

        data = self.__fetch(crawler_task=crawler_task, limit=limit, last_crawled=last_crawled)

        return data



from datetime import timedelta, datetime, timezone
from logging import getLogger
from typing import Optional

from flask import current_app
from flask_sqlalchemy import SQLAlchemy

from avatar_utils.sso_helper import SSOHelper

USER_ONLINE_STATE_TTL = timedelta(seconds=10)

logger = getLogger(__name__)


class SSOUserMixin:
    def __new__(cls, db: SQLAlchemy):
        fields = dict(
            sso_id=db.Column(db.String, nullable=False, index=True),
            latest_session_started_at=db.Column(db.DateTime(timezone=True), nullable=True, default=None),
            latest_session_request_at=db.Column(db.DateTime(timezone=True), nullable=True, default=None),
        )
        return type('_SSOUserMixin', (SSOUserBase,), fields)


class SSOUserBase:
    sso_id = None
    latest_session_started_at = None
    latest_session_request_at = None

    @classmethod
    def auth_callback(cls: type,
                      token: str):

        sso_client_id = current_app.config.get('SSO_CLIENT_ID')

        is_verify = current_app.config.get('SSO_VERIFY_TOKEN', True)

        sso_helper = SSOHelper(client_id=sso_client_id)
        payload = sso_helper.extract_payload(token=token, verify=is_verify)

        if payload is None or 'sub' not in payload:
            return None

        sso_id = payload['sub']
        user = cls.ensure(sso_id=sso_id)  # pylint: disable=no-member
        user.update(latest_session_request_at=datetime.utcnow().replace(tzinfo=timezone.utc))

        return user

    @property
    def is_online(self) -> Optional[bool]:
        online_state: Optional[bool] = None
        if self.latest_session_request_at:
            time_delta: timedelta = datetime.utcnow() - self.latest_session_request_at
            online_state = USER_ONLINE_STATE_TTL >= time_delta
        return online_state

    @property
    def latest_session_duration(self) -> Optional[timedelta]:
        if not self.latest_session_request_at and not self.latest_session_started_at:
            return None
        session_duration = self.latest_session_request_at - self.latest_session_started_at
        if not self.is_online:
            session_duration += USER_ONLINE_STATE_TTL
        return session_duration

    def update_latest_session_time(self) -> datetime:
        timestamp = datetime.utcnow().replace(tzinfo=timezone.utc)
        if not self.is_online:
            logger.info(f'User {self.sso_id} came back')
            self.update(latest_session_started_at=timestamp)  # pylint: disable=no-member
        self.update(latest_session_request_at=timestamp)  # pylint: disable=no-member
        return timestamp



from avatar_utils.db.mixins.created_at import CreatedAtMixin
from avatar_utils.db.mixins.crud import CRUDMixin
from avatar_utils.db.mixins.json_cast import JsonCastMixin
from avatar_utils.db.mixins.sso_user import SSOUserMixin
from avatar_utils.db.mixins.time_cast import TimeCastMixin

__all__ = [
    'CreatedAtMixin',
    'CRUDMixin',
    'JsonCastMixin',
    'TimeCastMixin',
    'SSOUserMixin',
]



from typing import Optional, Any, List, TypeVar

from app import db


class CRUDMixin:

    @classmethod
    def ensure(cls,
               **kwargs) -> db.Model:
        instance = cls.read(**kwargs)
        return instance if instance is not None else cls.create(**kwargs)

    @classmethod
    def create(cls, **kwargs) -> 'CRUDMixin':
        kwargs = cls._preprocess_params(**kwargs)
        instance = cls(**kwargs)  # noqa
        db.session.add(instance)
        db.session.commit()
        return instance

    @classmethod
    def all(cls, **kwargs) -> Optional[List['CRUDMixin']]:
        kwargs = cls._preprocess_params(**kwargs)
        return cls.query.filter_by(**kwargs).all()  # noqa

    @classmethod
    def read(cls, **kwargs) -> Optional['CRUDMixin']:
        kwargs = cls._preprocess_params(**kwargs)
        return cls.query.filter_by(**kwargs).first()  # noqa

    def update(self, **kwargs):
        kwargs = self._preprocess_params(**kwargs)
        for attr, value in kwargs.items():
            setattr(self, attr, value)
        db.session.commit()

    def save(self):
        db.session.add(self)
        db.session.flush()

    @classmethod
    def delete(cls, **kwargs):
        kwargs = cls._preprocess_params(**kwargs)
        rec = cls.read(**kwargs)
        db.session.delete(rec)
        db.session.commit()

    @staticmethod
    def _preprocess_params(**kwargs: Any) -> dict:
        return kwargs



import copy
import logging
from unittest import TestCase

from avatar_utils.tests.RemoteApp import RemoteApp

logger = logging.getLogger(__name__)


class IntegrationTest(TestCase, RemoteApp):
    logger.handlers.clear()
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(levelname)s %(message)s')

    def setUp(self):
        super().setUp()

        # reset current env
        self.current_env = copy.deepcopy(RemoteApp.default_env)



from .AbstractApp import AbstractApp
from .Env import *
from .RemoteApp import RemoteApp
from .IntegrationTest import IntegrationTest
from .UnitTest import UnitTest, TestCase


__all__ = ['AbstractApp', 'TestCase', 'Env','DevEnv','TestEnv','ProdEnv','RemoteApp', 'IntegrationTest','UnitTest']



import logging

from avatar_utils.sso_helper import SSOHelper
from avatar_utils.tests.UnitTest import UnitTest

logger = logging.getLogger(__name__)

sso = SSOHelper(client_id='client_id')


class BaseUnitTest(UnitTest):
    from app import create_app, db

    app = create_app('testing')
    client = app.test_client()
    db = db

    sso_id = 'test'

    token = None

    def setUp(self):
        super().setUp()

        self.token = sso.make_token(sso_user_id=self.sso_id)

        auth_header = {'Authorization': f'Bearer {self.token}'}
        self.headers = {**self.headers, **auth_header}



class LazyCachedAttribute(object):
    '''Computes attribute value and caches it in instance.

    Example:
        class MyClass(object):
            def myMethod(self):
                # ...
            myMethod = CachedAttribute(myMethod)
    Use "del inst.myMethod" to clear cache.'''

    def __init__(self, method, name: str = None, **method_args):
        self.method = method
        self.method_args=method_args
        self.name = name or method.__name__

    def __get__(self, inst, cls):
        if inst is None:
            return self
        result = self.method(inst, **self.method_args)
        setattr(inst, self.name, result)
        return result



from logging import getLogger
from typing import Optional, Dict

import requests
from flask import current_app

from avatar_utils.core.lazy_cached_attribute import LazyCachedAttribute

logger = getLogger()


class PlannerAdapter:

    def _make_url(self, rest):
        base_url = current_app.config["PLANNER_URL"]
        return f'{base_url}{rest}'

    _schedule_url = LazyCachedAttribute(method=_make_url, name='_schedule_url', rest='/schedule')
    _unschedule_url = LazyCachedAttribute(method=_make_url, name='_unschedule_url', rest='/unschedule')
    _status_url = LazyCachedAttribute(method=_make_url, name='_status_url', rest='/status')

    def schedule_route_call(self, **kwargs) -> Optional[requests.Response]:
        kwargs['service_name'] = kwargs.get('service_name') or current_app.config["SERVICE_NAME"]
        kwargs['method'] = kwargs.get('method') or 'POST'

        return self._call_http(url=self._schedule_url, json=kwargs)

    def unschedule_route_call(self, **kwargs) -> Optional[requests.Response]:
        kwargs['service_name'] = kwargs.get('service_name') or current_app.config["SERVICE_NAME"]
        kwargs['method'] = kwargs.get('method') or 'POST'

        return self._call_http(url=self._unschedule_url, json=kwargs)

    def status(self, **kwargs) -> Optional[requests.Response]:
        return self._call_http(url=self._unschedule_url, json=kwargs)

    def _call_http(self, url: str, json: dict = None, **kwargs) -> Optional[Dict]:
        response = requests.post(url=url, json=json, **kwargs)

        if response.status_code != 200:
            logger.warning(f'Something went wrong, status code: {response.status_code}. Response: {response.text}')

        return response.json()


planner_adapter = PlannerAdapter()
